/******************************************************************

 Copyright 2001 by Object Craft P/L, Melbourne, Australia.

 LICENCE - see LICENCE file distributed with this software for details.

******************************************************************/

#include "mssqldb.h"

static int bind_from_vartype(int vartype)
{
    switch (vartype) {
    case SQLBINARY: return VARYCHARBIND;
    case SQLCHAR: return VARYCHARBIND;
    case SQLBIT: return BITBIND;
    case SQLINT1: return TINYBIND;
    case SQLINT2: return SMALLBIND;
    case SQLINT4: return INTBIND;
    case SQLMONEY4: /* promote */
    case SQLMONEY: return MONEYBIND;
    case SQLFLT4: return FLT4BIND;
    case SQLFLT8: return FLT8BIND;
    case SQLDATETIM4: /* promote */
    case SQLDATETIME: return DATETIMEBIND;
    case SQLNUMERIC: return NUMERICBIND;
    case SQLDECIMAL: return DECIMALBIND;
    default: return vartype;
    }
}

static DBPROCObj *dbproc_list;

static void dbproc_add_object(DBPROCObj *dbproc)
{
    dbproc->next = dbproc_list;
    dbproc_list = dbproc;
}

static void dbproc_del_object(DBPROCObj *dbproc)
{
    DBPROCObj *scan, *prev;

    for (prev = NULL, scan = dbproc_list; scan != NULL; scan = scan->next) {
	if (scan == dbproc) {
	    if (prev == NULL)
		dbproc_list = scan->next;
	    else
		prev->next = scan->next;
	}
    }
}

PyObject *dbproc_find_object(DBPROCESS *db_dbproc)
{
    DBPROCObj *dbproc;

    for (dbproc = dbproc_list; dbproc != NULL; dbproc = dbproc->next)
	if (dbproc->dbproc == db_dbproc)
	    return (PyObject*)dbproc;
    return NULL;
}

static int has_been_freed(DBPROCObj *self)
{
    if (self->dbproc == NULL) {
        PyErr_SetString(PyExc_TypeError, "DBPROC has been freed");
	return 1;
    }
    return 0;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

typedef RETCODE (*DbProcFunc_retcode)(DBPROCESS *dbproc);

static PyObject *wrap_DbProcFunc_retcode(DBPROCObj *self, PyObject *args,
					 DbProcFunc_retcode func, char *name)
{
    RETCODE status;

    if (!PyArg_ParseTuple(args, "") || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = func(self->dbproc);
    if (self->debug)
	debug_msg("%s() -> %s\n", name, value_str(TYPE_RETCODE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

typedef DBBOOL (*DbProcFunc_bool)(DBPROCESS *dbproc);

static PyObject *wrap_DbProcFunc_bool(DBPROCObj *self, PyObject *args,
				      DbProcFunc_bool func, char *name)
{
    DBBOOL status;

    if (!PyArg_ParseTuple(args, "") || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = func(self->dbproc);
    if (self->debug)
	debug_msg("%s() -> %s\n", name, value_str(TYPE_RETCODE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

typedef DBINT (*DbProcFunc_dbint)(DBPROCESS *dbproc);

static PyObject *wrap_DbProcFunc_dbint(DBPROCObj *self, PyObject *args,
				       DbProcFunc_dbint func, char *name)
{
    DBINT status;

    if (!PyArg_ParseTuple(args, "") || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = func(self->dbproc);
    if (self->debug)
	debug_msg("%s() -> %d\n", name, status);
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

typedef int (*DbProcFunc_int)(DBPROCESS *dbproc);

static PyObject *wrap_DbProcFunc_int(DBPROCObj *self, PyObject *args,
				     DbProcFunc_int func, char *name)
{
    RETCODE status;

    if (!PyArg_ParseTuple(args, "") || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = func(self->dbproc);
    if (self->debug)
	debug_msg("%s() -> %d\n", name, status);
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

typedef void (*DbProcFunc)(DBPROCESS *dbproc);

static PyObject *wrap_DbProcFunc(DBPROCObj *self, PyObject *args,
				 DbProcFunc func, char *name)
{
    if (!PyArg_ParseTuple(args, "") || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    func(self->dbproc);
    if (self->debug)
	debug_msg("%s()\n", name);
    if (PyErr_Occurred())
	return NULL;

    Py_INCREF(Py_None);
    return Py_None;
}

typedef DBINT (*DbProcIntFunc_dbint)(DBPROCESS *dbproc, int num);

static PyObject *wrap_DbProcIntFunc_dbint(DBPROCObj *self, PyObject *args,
					  DbProcIntFunc_dbint func, char *name)
{
    int num;
    DBINT status;

    if (!PyArg_ParseTuple(args, "i", &num) || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = func(self->dbproc, num);
    if (self->debug)
	debug_msg("%s(%d) -> %d\n", name, num, status);
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

typedef DBBOOL (*DbProcIntFunc_bool)(DBPROCESS *dbproc, int num);

static PyObject *wrap_DbProcIntFunc_bool(DBPROCObj *self, PyObject *args,
					 DbProcIntFunc_bool func, char *name)
{
    int num;
    DBBOOL status;

    if (!PyArg_ParseTuple(args, "i", &num) || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = func(self->dbproc, num);
    if (self->debug)
	debug_msg("%s(%d) -> %s\n", name, num, value_str(TYPE_RETCODE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

typedef int (*DbProcIntFunc_int)(DBPROCESS *dbproc, int num);

static PyObject *wrap_DbProcIntFunc_type(DBPROCObj *self, PyObject *args,
					 DbProcIntFunc_int func, char *name)
{
    int num;
    int status;

    if (!PyArg_ParseTuple(args, "i", &num) || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = func(self->dbproc, num);
    if (self->debug)
	debug_msg("%s(%d) -> %s\n", name, num, value_str(TYPE_SERVTYPE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

static PyObject *wrap_DbProcIntFunc_int(DBPROCObj *self, PyObject *args,
					DbProcIntFunc_int func, char *name)
{
    int num;
    int status;

    if (!PyArg_ParseTuple(args, "i", &num) || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = func(self->dbproc, num);
    if (self->debug)
	debug_msg("%s(%d) -> %d\n", name, num, status);
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

#ifdef _WIN32
typedef const char *(*DbProcIntFunc_str)(DBPROCESS *dbproc, int num);
#else
typedef char *(*DbProcIntFunc_str)(DBPROCESS *dbproc, int num);
#endif

static PyObject *wrap_DbProcIntFunc_str(DBPROCObj *self, PyObject *args,
					DbProcIntFunc_str func, char *name)
{
    int num;
    const char *str;

    if (!PyArg_ParseTuple(args, "i", &num) || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    str = func(self->dbproc, num);
    if (self->debug) {
	if (str == NULL)
	    debug_msg("%s(%d) -> NULL\n", name, num);
	else
	    debug_msg("%s(%d) -> '%s'\n", name, num, str);
    }
    if (PyErr_Occurred())
	return NULL;

    if (str != NULL)
	return PyString_FromString(str);
    Py_INCREF(Py_None);
    return Py_None;
}

typedef int (*DbProcIntIntFunc_int)(DBPROCESS *dbproc, int num1, int num2);

static PyObject *wrap_DbProcIntIntFunc_int(DBPROCObj *self, PyObject *args,
					   DbProcIntIntFunc_int func,
					   char *name)
{
    int num1, num2;
    int status;

    if (!PyArg_ParseTuple(args, "ii", &num1, &num2) || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = func(self->dbproc, num1, num2);
    if (self->debug)
	debug_msg("%s(%d, %d) -> %d\n", name, num1, num2, status);
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

static PyObject *wrap_DbProcIntIntFunc_type(DBPROCObj *self, PyObject *args,
					    DbProcIntIntFunc_int func,
					    char *name)
{
    int num1, num2;
    int status;

    if (!PyArg_ParseTuple(args, "ii", &num1, &num2) || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = func(self->dbproc, num1, num2);
    if (self->debug)
	debug_msg("%s(%d, %d) -> %s\n", name, num1, num2, value_str(TYPE_SERVTYPE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

typedef DBINT (*DbProcIntIntFunc_dbint)(DBPROCESS *dbproc, int num1, int num2);

static PyObject *wrap_DbProcIntIntFunc_dbint(DBPROCObj *self, PyObject *args,
					     DbProcIntIntFunc_dbint func,
					     char *name)
{
    int num1, num2;
    DBINT len;

    if (!PyArg_ParseTuple(args, "ii", &num1, &num2) || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    len = func(self->dbproc, num1, num2);
    if (self->debug)
	debug_msg("%s(%d, %d) -> %d\n", name, num1, num2, len);
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(len);
}

#ifdef _WIN32
typedef LPCBYTE (*DbProcDataFunc)(PDBPROCESS dbproc, INT num);
#else
typedef BYTE *(*DbProcDataFunc)(DBPROCESS *dbproc, int num);
#endif
typedef DBINT (*DbProcDataLenFunc)(DBPROCESS *dbproc, int num);

static PyObject *wrap_DbData(DBPROCObj *self, PyObject *args,
			     DbProcDataFunc data_func,
			     DbProcDataLenFunc datalen_func,
			     char *data_name, char *len_name)
{
    DataBufObj *databuf;
    int num, len;
    const BYTE *ptr;
    RETCODE status;

    len = -1;
    if (!PyArg_ParseTuple(args, "O!i|i",
			  &DataBufType, &databuf, &num, &len)
	|| has_been_freed(self))
	return NULL;

    status = SUCCEED;
    if (len < 0)
	len = databuf->varlen;

    PyErr_Clear();
    ptr = data_func(self->dbproc, num);
    if (self->debug) {
	if (ptr == NULL)
	    debug_msg("%s(%d) -> NULL\n", data_name, num);
	else
	    debug_msg("%s(%d) -> %p\n", data_name, num, ptr);
    }
    if (ptr == NULL) {
	int res;

	res = datalen_func(self->dbproc, num);
	if (self->debug)
	    debug_msg("%s(%d) -> %d\n", len_name, num, res);
	if (res == 0)
	    databuf->varlen = 0;
	else
	    status = FAIL;
    } else {
	memmove(databuf->buff, ptr, len);
	if (len < databuf->size)
	    databuf->buff[len] = '\0';
    }
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static char DBPROC_dbadata__doc__[] = 
"dbadata(databuf, computeid, column [, len]) -> bool";

static PyObject *DBPROC_dbadata(DBPROCObj *self, PyObject *args)
{
    DataBufObj *databuf;
    int computeid, num, len;
    const BYTE *ptr;
    RETCODE status;

    len = -1;
    if (!PyArg_ParseTuple(args, "O!ii|i",
			  &DataBufType, &databuf, &computeid, &num, &len)
	|| has_been_freed(self))
	return NULL;

    status = SUCCEED;
    if (len < 0)
	len = databuf->varlen;

    PyErr_Clear();
    ptr = dbadata(self->dbproc, computeid, num);
    if (self->debug) {
	if (ptr == NULL)
	    debug_msg("dbadata(%d, %d) -> NULL\n", computeid, num);
	else
	    debug_msg("dbadata(%d, %d) -> %p\n", computeid, num, ptr);
    }
    if (ptr == NULL) {
	int res;

	res = dbaltlen(self->dbproc, computeid, num);
	if (self->debug)
	    debug_msg("dbaltlen(%d, %d) -> %d\n", computeid, num, res);
	if (res == 0)
	    databuf->varlen = 0;
	else
	    status = FAIL;
    } else {
	memmove(databuf->buff, ptr, len);
	if (len < databuf->size)
	    databuf->buff[len] = '\0';
    }
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

static char DBPROC_dbadlen__doc__[] = 
"dbadlen(computeid, column) -> int";

static PyObject *DBPROC_dbadlen(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntIntFunc_dbint(self, args, dbadlen, "dbadlen");
}

#ifdef HAVE_DBALTBIND
static char DBPROC_dbaltbind__doc__[] = 
"dbaltbind(computeid, column, databuf [, len]) -> bool";

static PyObject *DBPROC_dbaltbind(DBPROCObj *self, PyObject *args)
{
    int computeid, column, len;
    DataBufObj *buf;
    RETCODE status;

    len = -1;
    if (!PyArg_ParseTuple(args, "iiO!|i",
			  &computeid, &column, &DataBufType, &buf, &len)
	|| has_been_freed(self))
	return NULL;

    if (len < 0)
	len = buf->varlen;

    PyErr_Clear();
    status = dbaltbind(self->dbproc, computeid, column,
		       bind_from_vartype(buf->vartype), len, buf->buff);
    if (self->debug)
	debug_msg("dbaltbind(%d, %d, %s, %d, %p) -> %s\n",
		  computeid, column,
		  value_str(TYPE_BIND, bind_from_vartype(buf->vartype)), len,
		  buf->buff, value_str(TYPE_RETCODE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}
#endif

static char DBPROC_dbaltcolid__doc__[] = 
"dbaltcolid(computeid, column) -> int";

static PyObject *DBPROC_dbaltcolid(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntIntFunc_int(self, args, dbaltcolid, "dbaltcolid");
}

static char DBPROC_dbalttype__doc__[] = 
"dbalttype(computeid, column) -> int";

static PyObject *DBPROC_dbalttype(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntIntFunc_type(self, args, dbalttype, "dbalttype");
}

static char DBPROC_dbaltlen__doc__[] = 
"dbaltlen(computeid, column) -> int";

static PyObject *DBPROC_dbaltlen(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntIntFunc_dbint(self, args, dbaltlen, "dbaltlen");
}

static char DBPROC_dbbind__doc__[] = 
"dbbind(column, databuf [, len]) -> bool";

static PyObject *DBPROC_dbbind(DBPROCObj *self, PyObject *args)
{
    int column, len;
    DataBufObj *buf;
    RETCODE status;

    len = -1;
    if (!PyArg_ParseTuple(args, "iO!|i", &column, &DataBufType, &buf, &len)
	|| has_been_freed(self))
	return NULL;

    if (len < 0)
	len = buf->varlen;

    PyErr_Clear();
    status = dbbind(self->dbproc, column,
		    bind_from_vartype(buf->vartype), len, buf->buff);
    if (self->debug)
	debug_msg("dbbind(%d, %s, %d, %p) -> %s\n",
		  column,
		  value_str(TYPE_BIND, bind_from_vartype(buf->vartype)), len,
		  buf->buff, value_str(TYPE_RETCODE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

static char DBPROC_dbcancel__doc__[] = 
"dbcancel() -> bool";

static PyObject *DBPROC_dbcancel(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_retcode(self, args, dbcancel, "dbcancel");
}

static char DBPROC_dbcanquery__doc__[] = 
"dbcanquery() -> bool";

static PyObject *DBPROC_dbcanquery(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_retcode(self, args, dbcanquery, "dbcanquery");
}

static char DBPROC_dbclose__doc__[] = 
"dbclose() -> None";

static PyObject *DBPROC_dbclose(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc(self, args, (DbProcFunc)dbclose, "dbclose");
}

#ifdef HAVE_DBCLROPT
static char DBPROC_dbclropt__doc__[] = 
"dbclropt(option [, param]) -> bool";

static PyObject *DBPROC_dbclropt(DBPROCObj *self, PyObject *args)
{
    int option;
    char *param = NULL;
    RETCODE status;

    if (!PyArg_ParseTuple(args, "i|s", &option, &param)
	|| has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = dbclropt(self->dbproc, option, param);
    if (self->debug) {
	if (param == NULL)
	    debug_msg("dbclropt(%s, NULL) -> %s\n",
		      value_str(TYPE_OPTION, option),
		      value_str(TYPE_RETCODE, status));
	else
	    debug_msg("dbclropt(%s, '%s') -> %s\n",
		      value_str(TYPE_OPTION, option), param,
		      value_str(TYPE_RETCODE, status));
    }
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}
#endif

static char DBPROC_dbcmd__doc__[] = 
"dbcmd(str) -> bool";

static PyObject *DBPROC_dbcmd(DBPROCObj *self, PyObject *args)
{
    char *cmd;
    RETCODE status;

    if (!PyArg_ParseTuple(args, "s", &cmd) || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = dbcmd(self->dbproc, cmd);
    if (self->debug)
	debug_msg("dbcmd('%s') -> %s\n", cmd, value_str(TYPE_RETCODE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

static char DBPROC_dbcmdrow__doc__[] = 
"dbcmdrow() -> bool";

static PyObject *DBPROC_dbcmdrow(DBPROCObj *self, PyObject *args)
{
    RETCODE status;

    if (!PyArg_ParseTuple(args, "") || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = DBCMDROW(self->dbproc);
    if (self->debug)
	debug_msg("DBCMDROW() -> %s\n", value_str(TYPE_RETCODE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

#ifdef HAVE_DBCOLBROWSE
static char DBPROC_dbcolbrowse__doc__[] = 
"dbcolbrowse(column) -> bool";

static PyObject *DBPROC_dbcolbrowse(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_bool(self, args, dbcolbrowse, "dbcolbrowse");
}
#endif

static char DBPROC_dbcollen__doc__[] = 
"dbcollen(column) -> len";

static PyObject *DBPROC_dbcollen(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_dbint(self, args, dbcollen, "dbcollen");
}

static char DBPROC_dbcolname__doc__[] = 
"dbcolname(column) -> name";

static PyObject *DBPROC_dbcolname(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_str(self, args, dbcolname, "dbcolname");
}

static char DBPROC_dbcolsource__doc__[] = 
"dbcolsource(column) -> name";

static PyObject *DBPROC_dbcolsource(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_str(self, args, dbcolsource, "dbcolsource");
}

static char DBPROC_dbcoltype__doc__[] = 
"dbcoltype(column) -> int";

static PyObject *DBPROC_dbcoltype(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_type(self, args, dbcoltype, "dbcoltype");
}

static char DBPROC_dbconvert__doc__[] = 
"dbconvert(obj, databuf) -> len";

static PyObject *DBPROC_dbconvert(DBPROCObj *self, PyObject *args)
{
    PyObject *from;
    DataBufObj *to;

    if (!PyArg_ParseTuple(args, "OO!", &from, &DataBufType, &to)
	|| has_been_freed(self))
	return NULL;
    return DataBuf_convert(self, from, to);
}

#ifdef HAVE_DBCURCMD
static char DBPROC_dbcurcmd__doc__[] = 
"dbcurcmd() -> int";

static PyObject *DBPROC_dbcurcmd(DBPROCObj *self, PyObject *args)
{
    int num;

    if (!PyArg_ParseTuple(args, "") || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    num = DBCURCMD(self->dbproc);
    if (self->debug)
	debug_msg("DBCURCMD() -> %d\n", num);
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(num);
}
#endif

#ifdef HAVE_DBCURROW
static char DBPROC_dbcurrow__doc__[] = 
"dbcurrow() -> int";

static PyObject *DBPROC_dbcurrow(DBPROCObj *self, PyObject *args)
{
    int num;

    if (!PyArg_ParseTuple(args, "") || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    num = DBCURROW(self->dbproc);
    if (self->debug)
	debug_msg("DBCURROW() -> %d\n", num);
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(num);
}
#endif

static char DBPROC_dbdata__doc__[] = 
"dbdata(databuf, column [, len]) -> bool";

static PyObject *DBPROC_dbdata(DBPROCObj *self, PyObject *args)
{
    return wrap_DbData(self, args, dbdata, dbdatlen, "dbdata", "dbdatlen");
}

static char DBPROC_dbdatlen__doc__[] = 
"dbdatlen(column) -> int";

static PyObject *DBPROC_dbdatlen(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_dbint(self, args, dbdatlen, "dbdatlen");
}

static char DBPROC_dbdead__doc__[] = 
"dbdead() -> bool";

static PyObject *DBPROC_dbdead(DBPROCObj *self, PyObject *args)
{
    RETCODE status;

    if (!PyArg_ParseTuple(args, "") || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = DBDEAD(self->dbproc);
    if (self->debug)
	debug_msg("DBDEAD() -> %s\n", value_str(TYPE_RETCODE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

static char DBPROC_dbhasretstat__doc__[] = 
"dbhasretstat() -> bool";

static PyObject *DBPROC_dbhasretstat(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_bool(self, args, dbhasretstat, "dbhasretstat");
}

#ifdef HAVE_DBISOPT
static char DBPROC_dbisopt__doc__[] = 
"dbisopt(option [, param]) -> bool";

static PyObject *DBPROC_dbisopt(DBPROCObj *self, PyObject *args)
{
    int option;
    char *param = NULL;
    RETCODE status;

    if (!PyArg_ParseTuple(args, "i|s", &option, &param)
	|| has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = dbisopt(self->dbproc, option, param);
    if (self->debug) {
	if (param == NULL)
	    debug_msg("dbisopt(%s, NULL) -> %s\n",
		      value_str(TYPE_OPTION, option),
		      value_str(TYPE_RETCODE, status));
	else
	    debug_msg("dbisopt(%s, '%s') -> %s\n",
		      value_str(TYPE_OPTION, option), param,
		      value_str(TYPE_RETCODE, status));
    }
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}
#endif

static char DBPROC_dbnextrow__doc__[] = 
"dbnextrow() -> status";

static PyObject *DBPROC_dbnextrow(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_retcode(self, args, dbnextrow, "dbnextrow");
}

static char DBPROC_dbnumalts__doc__[] = 
"dbnumalts(computeid) -> int";

static PyObject *DBPROC_dbnumalts(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_int(self, args, dbnumalts, "dbnumalts");
}

static char DBPROC_dbnumcols__doc__[] = 
"dbnumcols() -> int";

static PyObject *DBPROC_dbnumcols(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_int(self, args, dbnumcols, "dbnumcols");
}

#ifdef HAVE_DBNUMCOMPUTE
static char DBPROC_dbnumcompute__doc__[] = 
"dbnumcompute() -> int";

static PyObject *DBPROC_dbnumcompute(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_retcode(self, args, dbnumcompute, "dbnumcompute");
}
#endif

static char DBPROC_dbnumrets__doc__[] = 
"dbnumrets() -> int";

static PyObject *DBPROC_dbnumrets(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_retcode(self, args, dbnumrets, "dbnumrets");
}

static char DBPROC_dbprhead__doc__[] = 
"dbprhead() -> None";

static PyObject *DBPROC_dbprhead(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc(self, args, dbprhead, "dbprhead");
}

#ifdef HAVE_DBPRROW
static char DBPROC_dbprrow__doc__[] = 
"dbprrow() -> bool";

static PyObject *DBPROC_dbprrow(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_retcode(self, args, (DbProcFunc_retcode)dbprrow, "dbprrow");
}
#endif

static char DBPROC_dbresults__doc__[] = 
"dbresults() -> status";

static PyObject *DBPROC_dbresults(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_retcode(self, args, dbresults, "dbresults");
}

static char DBPROC_dbretdata__doc__[] = 
"dbretdata(databuf, retnum [, len]) -> bool";

static PyObject *DBPROC_dbretdata(DBPROCObj *self, PyObject *args)
{
    return wrap_DbData(self, args, dbretdata, dbretlen, "dbretdata", "dbretlen");
}

static char DBPROC_dbretlen__doc__[] = 
"dbretlen(retnum) -> int";

static PyObject *DBPROC_dbretlen(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_dbint(self, args, dbretlen, "dbretlen");
}

static char DBPROC_dbretname__doc__[] = 
"dbretname(retnum) -> name";

static PyObject *DBPROC_dbretname(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_str(self, args, dbretname, "dbretname");
}

static char DBPROC_dbretstatus__doc__[] = 
"dbretstatus() -> int";

static PyObject *DBPROC_dbretstatus(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_dbint(self, args, dbretstatus, "dbretstatus");
}

#ifdef HAVE_DBRETTYPE
static char DBPROC_dbrettype__doc__[] = 
"dbrettype(retnum) -> int";

static PyObject *DBPROC_dbrettype(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_type(self, args, dbrettype, "dbrettype");
}
#endif

static char DBPROC_dbrpcinit__doc__[] = 
"dbrpcinit(rpcname [, options = 0]) -> bool";

static PyObject *DBPROC_dbrpcinit(DBPROCObj *self, PyObject *args)
{
    char *rpcname;
    int options = 0;
    RETCODE status;

    if (!PyArg_ParseTuple(args, "s|i", &rpcname, &options)
	|| has_been_freed(self))
	return NULL;

    PyErr_Clear();
#ifdef _WIN32
    status = dbrpcinit(self->dbproc, rpcname, (DBSMALLINT)options);
#else
    status = dbrpcinit(self->dbproc, rpcname, options);
#endif
    if (self->debug)
	debug_msg("dbrpcinit('%s', %d) -> %s\n", rpcname, options,
		  value_str(TYPE_RETCODE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

static char DBPROC_dbrpcparam__doc__[] = 
"dbrpcparam(paramname, databuf [, status = 0]) -> bool";

static PyObject *DBPROC_dbrpcparam(DBPROCObj *self, PyObject *args)
{
    char *paramname;
    DataBufObj *databuf;
    int status = 0, maxlen, datalen;
    RETCODE retcode;

    if (!PyArg_ParseTuple(args, "sO!|i",
			  &paramname, &DataBufType, &databuf, &status)
	|| has_been_freed(self))
	return NULL;

    if (status & DBRPCRETURN)
	maxlen = -1;
    else
	maxlen = databuf->size;
    if (is_fixedlen(databuf->vartype))
	datalen = -1;
    else
	datalen = databuf->varlen;
    PyErr_Clear();
#ifdef _WIN32
    retcode = dbrpcparam(self->dbproc, paramname, (BYTE)status,
			 databuf->vartype, maxlen, datalen, databuf->buff);
#else
    retcode = dbrpcparam(self->dbproc, paramname, status,
			 databuf->vartype, maxlen, datalen, databuf->buff);
#endif
    if (self->debug)
	debug_msg("dbrpcparam('%s', %d, %s, %d, %d, %p) -> %s\n",
		  paramname, status,
		  value_str(TYPE_SERVTYPE, status), maxlen, datalen,
		  databuf->buff, value_str(TYPE_RETCODE, retcode));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(retcode);
}

static char DBPROC_dbrpcsend__doc__[] = 
"dbrpcsend() -> bool";

static PyObject *DBPROC_dbrpcsend(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_retcode(self, args, dbrpcsend, "dbrpcsend");
}

static char DBPROC_dbsetopt__doc__[] = 
"dbsetopt(option [, name [, value]]) -> bool";

static PyObject *DBPROC_dbsetopt(DBPROCObj *self, PyObject *args)
{
    int option;
    char *char_param = NULL;
    int int_param = -1;
    RETCODE retcode;

    if (!PyArg_ParseTuple(args, "i|si", &option, &char_param, &int_param)
	|| has_been_freed(self))
	return NULL;

    PyErr_Clear();
#ifdef _WIN32
    retcode = dbsetopt(self->dbproc, option, char_param);
    if (self->debug) {
	if (char_param == NULL)
	    debug_msg("dbsetopt(%s, NULL) -> %s\n",
		      value_str(TYPE_OPTION, option),
		      value_str(TYPE_RETCODE, retcode));
	else
	    debug_msg("dbsetopt(%s, '%s') -> %s\n",
		      value_str(TYPE_OPTION, option), char_param,
		      value_str(TYPE_RETCODE, retcode));
    }
#else
    retcode = dbsetopt(self->dbproc, option, char_param, int_param);
    if (self->debug) {
	if (char_param == NULL)
	    debug_msg("dbsetopt(%s, NULL, %d) -> %s\n",
		      value_str(TYPE_OPTION, option), int_param,
		      value_str(TYPE_RETCODE, retcode));
	else
	    debug_msg("dbsetopt(%s, '%s', %d) -> %s\n",
		      value_str(TYPE_OPTION, option), char_param, int_param,
		      value_str(TYPE_RETCODE, retcode));
    }
#endif
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(retcode);
}

static char DBPROC_dbsqlexec__doc__[] = 
"dbsqlexec() -> bool";

static PyObject *DBPROC_dbsqlexec(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_retcode(self, args, dbsqlexec, "dbsqlexec");
}

static char DBPROC_dbsqlok__doc__[] = 
"dbsqlok() -> bool";

static PyObject *DBPROC_dbsqlok(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_retcode(self, args, dbsqlok, "dbsqlok");
}

#ifdef HAVE_DBTABBROWSE
static char DBPROC_dbtabbrowse__doc__[] = 
"dbtabbrowse(tabnum) -> bool";

static PyObject *DBPROC_dbtabbrowse(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_bool(self, args, dbtabbrowse, "dbtabbrowse");
}
#endif

#ifdef HAVE_DBTABCOUNT
static char DBPROC_dbtabcount__doc__[] = 
"dbtabcount(tabnum) -> int";

static PyObject *DBPROC_dbtabcount(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcFunc_retcode(self, args, dbtabcount, "dbtabcount");
}
#endif

#ifdef HAVE_DBTABNAME
static char DBPROC_dbtabname__doc__[] = 
"dbtabname(tabnum) -> name";

static PyObject *DBPROC_dbtabname(DBPROCObj *self, PyObject *args)
{
    return wrap_DbProcIntFunc_str(self, args, dbtabname, "dbtabname");
}
#endif

#ifdef HAVE_DBTABSOURCE
static char DBPROC_dbtabsource__doc__[] = 
"dbtabsource(column) -> name, tabnum";

static PyObject *DBPROC_dbtabsource(DBPROCObj *self, PyObject *args)
{
    int column, tabnum;
    const char *tabname;

    if (!PyArg_ParseTuple(args, "i", &column) || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    tabname = dbtabsource(self->dbproc, column, &tabnum);
    if (self->debug) {
	if (tabname == NULL)
	    debug_msg("dbtabsource(%d) -> NULL, %d\n", column, tabnum);
	else
	    debug_msg("dbtabsource(%d) -> '%s', %d\n", column, tabname, tabnum);
    }
    if (PyErr_Occurred())
	return NULL;

    if (tabname == NULL)
	return Py_BuildValue("Oi", Py_None, tabnum);
    return Py_BuildValue("si", tabname, tabnum);
}
#endif

static char DBPROC_dbuse__doc__[] = 
"dbuse(dbname) -> bool";

static PyObject *DBPROC_dbuse(DBPROCObj *self, PyObject *args)
{
    char *dbname;
    RETCODE status;

    if (!PyArg_ParseTuple(args, "s", &dbname) || has_been_freed(self))
	return NULL;

    PyErr_Clear();
    status = dbuse(self->dbproc, dbname);
    if (self->debug)
	debug_msg("dbuse('%s') -> %s\n",
		  dbname, value_str(TYPE_RETCODE, status));
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}

static struct PyMethodDef DBPROC_methods[] = {
    { "dbadata", (PyCFunction)DBPROC_dbadata, METH_VARARGS, DBPROC_dbadata__doc__ },
    { "dbadlen", (PyCFunction)DBPROC_dbadlen, METH_VARARGS, DBPROC_dbadlen__doc__ },
#ifdef HAVE_DBALTBIND
    { "dbaltbind", (PyCFunction)DBPROC_dbaltbind, METH_VARARGS, DBPROC_dbaltbind__doc__ },
#endif
    { "dbaltcolid", (PyCFunction)DBPROC_dbaltcolid, METH_VARARGS, DBPROC_dbaltcolid__doc__ },
    { "dbaltlen", (PyCFunction)DBPROC_dbaltlen, METH_VARARGS, DBPROC_dbaltlen__doc__ },
    { "dbalttype", (PyCFunction)DBPROC_dbalttype, METH_VARARGS, DBPROC_dbalttype__doc__ },
    { "dbbind", (PyCFunction)DBPROC_dbbind, METH_VARARGS, DBPROC_dbbind__doc__ },
    { "dbcancel", (PyCFunction)DBPROC_dbcancel, METH_VARARGS, DBPROC_dbcancel__doc__ },
    { "dbcanquery", (PyCFunction)DBPROC_dbcanquery, METH_VARARGS, DBPROC_dbcanquery__doc__ },
    { "dbclose", (PyCFunction)DBPROC_dbclose, METH_VARARGS, DBPROC_dbclose__doc__ },
#ifdef HAVE_DBCLROPT
    { "dbclropt", (PyCFunction)DBPROC_dbclropt, METH_VARARGS, DBPROC_dbclropt__doc__ },
#endif
    { "dbcmd", (PyCFunction)DBPROC_dbcmd, METH_VARARGS, DBPROC_dbcmd__doc__ },
    { "dbcmdrow", (PyCFunction)DBPROC_dbcmdrow, METH_VARARGS, DBPROC_dbcmdrow__doc__ },
#ifdef HAVE_DBCOLBROWSE
    { "dbcolbrowse", (PyCFunction)DBPROC_dbcolbrowse, METH_VARARGS, DBPROC_dbcolbrowse__doc__ },
#endif
    { "dbcollen", (PyCFunction)DBPROC_dbcollen, METH_VARARGS, DBPROC_dbcollen__doc__ },
    { "dbcolname", (PyCFunction)DBPROC_dbcolname, METH_VARARGS, DBPROC_dbcolname__doc__ },
    { "dbcolsource", (PyCFunction)DBPROC_dbcolsource, METH_VARARGS, DBPROC_dbcolsource__doc__ },
    { "dbcoltype", (PyCFunction)DBPROC_dbcoltype, METH_VARARGS, DBPROC_dbcoltype__doc__ },
    { "dbconvert", (PyCFunction)DBPROC_dbconvert, METH_VARARGS, DBPROC_dbconvert__doc__ },
#ifdef HAVE_DBCURCMD
    { "dbcurcmd", (PyCFunction)DBPROC_dbcurcmd, METH_VARARGS, DBPROC_dbcurcmd__doc__ },
#endif
#ifdef HAVE_DBCURROW
    { "dbcurrow", (PyCFunction)DBPROC_dbcurrow, METH_VARARGS, DBPROC_dbcurrow__doc__ },
#endif
    { "dbdata", (PyCFunction)DBPROC_dbdata, METH_VARARGS, DBPROC_dbdata__doc__ },
    { "dbdatlen", (PyCFunction)DBPROC_dbdatlen, METH_VARARGS, DBPROC_dbdatlen__doc__ },
    { "dbdead", (PyCFunction)DBPROC_dbdead, METH_VARARGS, DBPROC_dbdead__doc__ },
    { "dbhasretstat", (PyCFunction)DBPROC_dbhasretstat, METH_VARARGS, DBPROC_dbhasretstat__doc__ },
#ifdef HAVE_DBISOPT
    { "dbisopt", (PyCFunction)DBPROC_dbisopt, METH_VARARGS, DBPROC_dbisopt__doc__ },
#endif
    { "dbnextrow", (PyCFunction)DBPROC_dbnextrow, METH_VARARGS, DBPROC_dbnextrow__doc__ },
    { "dbnumalts", (PyCFunction)DBPROC_dbnumalts, METH_VARARGS, DBPROC_dbnumalts__doc__ },
    { "dbnumcols", (PyCFunction)DBPROC_dbnumcols, METH_VARARGS, DBPROC_dbnumcols__doc__ },
#ifdef HAVE_DBNUMCOMPUTE
    { "dbnumcompute", (PyCFunction)DBPROC_dbnumcompute, METH_VARARGS, DBPROC_dbnumcompute__doc__ },
#endif
    { "dbnumrets", (PyCFunction)DBPROC_dbnumrets, METH_VARARGS, DBPROC_dbnumrets__doc__ },
    { "dbprhead", (PyCFunction)DBPROC_dbprhead, METH_VARARGS, DBPROC_dbprhead__doc__ },
#ifdef HAVE_DBPRROW
    { "dbprrow", (PyCFunction)DBPROC_dbprrow, METH_VARARGS, DBPROC_dbprrow__doc__ },
#endif
    { "dbresults", (PyCFunction)DBPROC_dbresults, METH_VARARGS, DBPROC_dbresults__doc__ },
    { "dbretdata", (PyCFunction)DBPROC_dbretdata, METH_VARARGS, DBPROC_dbretdata__doc__ },
    { "dbretlen", (PyCFunction)DBPROC_dbretlen, METH_VARARGS, DBPROC_dbretlen__doc__ },
    { "dbretname", (PyCFunction)DBPROC_dbretname, METH_VARARGS, DBPROC_dbretname__doc__ },
    { "dbretstatus", (PyCFunction)DBPROC_dbretstatus, METH_VARARGS, DBPROC_dbretstatus__doc__ },
#ifdef HAVE_DBRETTYPE
    { "dbrettype", (PyCFunction)DBPROC_dbrettype, METH_VARARGS, DBPROC_dbrettype__doc__ },
#endif
    { "dbrpcinit", (PyCFunction)DBPROC_dbrpcinit, METH_VARARGS, DBPROC_dbrpcinit__doc__ },
    { "dbrpcparam", (PyCFunction)DBPROC_dbrpcparam, METH_VARARGS, DBPROC_dbrpcparam__doc__ },
    { "dbrpcsend", (PyCFunction)DBPROC_dbrpcsend, METH_VARARGS, DBPROC_dbrpcsend__doc__ },
    { "dbsetopt", (PyCFunction)DBPROC_dbsetopt, METH_VARARGS, DBPROC_dbsetopt__doc__ },
    { "dbsqlexec", (PyCFunction)DBPROC_dbsqlexec, METH_VARARGS, DBPROC_dbsqlexec__doc__ },
    { "dbsqlok", (PyCFunction)DBPROC_dbsqlok, METH_VARARGS, DBPROC_dbsqlok__doc__ },
#ifdef HAVE_DBTABBROWSE
    { "dbtabbrowse", (PyCFunction)DBPROC_dbtabbrowse, METH_VARARGS, DBPROC_dbtabbrowse__doc__ },
#endif
#ifdef HAVE_DBTABCOUNT
    { "dbtabcount", (PyCFunction)DBPROC_dbtabcount, METH_VARARGS, DBPROC_dbtabcount__doc__ },
#endif
#ifdef HAVE_DBTABNAME
    { "dbtabname", (PyCFunction)DBPROC_dbtabname, METH_VARARGS, DBPROC_dbtabname__doc__ },
#endif
#ifdef HAVE_DBTABSOURCE
    { "dbtabsource", (PyCFunction)DBPROC_dbtabsource, METH_VARARGS, DBPROC_dbtabsource__doc__ },
#endif
    { "dbuse", (PyCFunction)DBPROC_dbuse, METH_VARARGS, DBPROC_dbuse__doc__ },
    { NULL, NULL }
};

PyObject *DBPROC__init__(LOGINREC *login, char *server, int debug)
{
    DBPROCESS *dbproc;
    DBPROCObj *self;

    dbproc = dbopen(login, server);
    if (debug) {
	if (dbproc == NULL)
	    debug_msg("dbopen('%s') -> FAILED\n", server);
	else
	    debug_msg("dbopen('%s') -> OK\n", server);
    }
    if (dbproc == NULL) {
	PyErr_SetString(PyExc_SystemError, "dbopen failed");
	return NULL;
    }

    self = PyObject_NEW(DBPROCObj, &DBPROCType);
    if (self == NULL)
	return NULL;
    self->dbproc = dbproc;
    self->debug = debug;
    dbproc_add_object(self);
    return (PyObject *)self;
}

static void DBPROC__del__(DBPROCObj *self)
{
    if (self->dbproc != NULL) {
	dbclose(self->dbproc);
	if (self->debug)
	    debug_msg("dbclose()\n");
    }

    dbproc_del_object(self);
    PyMem_DEL(self);
}

#define OFFSET(x) offsetof(DBPROCObj, x)

static struct memberlist DBPROC_memberlist[] = {
    { "debug", T_INT, OFFSET(debug) },
    { NULL }
};

static PyObject *DBPROC_getattr(DBPROCObj *self, char *name)
{
    PyObject *rv;

    rv = PyMember_Get((char *)self, DBPROC_memberlist, name);
    if (rv)
	return rv;
    PyErr_Clear();
    return Py_FindMethod(DBPROC_methods, (PyObject *)self, name);
}

static int DBPROC_setattr(DBPROCObj *self, char *name, PyObject *v)
{
    if (v == NULL) {
	PyErr_SetString(PyExc_AttributeError, "Cannot delete attribute");
	return -1;
    }
    return PyMember_Set((char *)self, DBPROC_memberlist, name, v);
}

static char DBPROCType__doc__[] = 
"";

PyTypeObject DBPROCType = {
    PyObject_HEAD_INIT(0)
    0,				/*ob_size*/
    "DBPROC",			/*tp_name*/
    sizeof(DBPROCObj),		/*tp_basicsize*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)DBPROC__del__,	/*tp_dealloc*/
    (printfunc)0,		/*tp_print*/
    (getattrfunc)DBPROC_getattr, /*tp_getattr*/
    (setattrfunc)DBPROC_setattr, /*tp_setattr*/
    (cmpfunc)0,			/*tp_compare*/
    (reprfunc)0,		/*tp_repr*/
    0,				/*tp_as_number*/
    0,				/*tp_as_sequence*/
    0,				/*tp_as_mapping*/
    (hashfunc)0,		/*tp_hash*/
    (ternaryfunc)0,		/*tp_call*/
    (reprfunc)0,		/*tp_str*/

    0L, 0L, 0L, 0L,
    DBPROCType__doc__
};
